/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is part of dcm4che, an implementation of DICOM(TM) in
 * Java(TM), hosted at https://github.com/dcm4che.
 *
 * The Initial Developer of the Original Code is
 * Agfa Healthcare.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * See @authors listed below
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

package com.teletip.makekos;

import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Date;
import java.util.Properties;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.dcm4che3.data.Attributes;
import org.dcm4che3.data.Sequence;
import org.dcm4che3.data.Tag;
import org.dcm4che3.data.UID;
import org.dcm4che3.data.VR;
import org.dcm4che3.io.DicomEncodingOptions;
import org.dcm4che3.io.DicomInputStream;
import org.dcm4che3.io.DicomOutputStream;
import org.dcm4che3.io.DicomInputStream.IncludeBulkData;
import org.dcm4che3.tool.common.CLIUtils;
import org.dcm4che3.tool.common.DicomFiles;
import org.dcm4che3.util.SafeClose;
import org.dcm4che3.util.UIDUtils;

public class MakeKOS {

    private static ResourceBundle rb =
        ResourceBundle.getBundle("com.teletip.makekos.messages");

    private static final int[] PATIENT_AND_STUDY_ATTRS = {
        Tag.SpecificCharacterSet,
        Tag.StudyDate,
        Tag.StudyTime,
        Tag.AccessionNumber,
        Tag.IssuerOfAccessionNumberSequence,
        Tag.ReferringPhysicianName,
        Tag.PatientName,
        Tag.PatientID,
        Tag.OtherPatientIDs,
        Tag.IssuerOfPatientID,
        Tag.PatientBirthDate,
        Tag.PatientSex,
        Tag.StudyInstanceUID,
        Tag.SOPInstanceUID,
        Tag.InstitutionName,
        Tag.StudyID,
        Tag.PixelData,
        Tag.DeviceSerialNumber
        };

    private final Attributes attrs = new Attributes();
    private String uidSuffix;
    private String fname;
    private boolean nofmi;
    private DicomEncodingOptions encOpts;
    private String tsuid;
    private String seriesNumber;
    private String instanceNumber;
    private String keyObjectDescription;
    private String retrieveAET;
    private String retrieveURL;
    private String locationUID;
    private String institutionName;
    private String deviceSerialNumber;
    private static String tempLocation;
    private static String dcmLocation;
    private Attributes documentTitle;
    private Attributes documentTitleModifier;
    private Properties codes;

    private Attributes kos;
    private Sequence evidenceSeq;
    private Sequence contentSeq;

    public String getFname() {
        return fname;
    }
        
    public final void setUIDSuffix(String uidSuffix) {
        this.uidSuffix = uidSuffix;
    }

    public void setOutputFile(String fname) {
        this.fname = fname;
    }

    public void setNoFileMetaInformation(boolean nofmi) {
        this.nofmi = nofmi;
    }

    public final void setEncodingOptions(DicomEncodingOptions encOpts) {
        this.encOpts = encOpts;
    }

    public final void setTransferSyntax(String tsuid) {
        this.tsuid = tsuid;
    }

    public final void setSeriesNumber(String seriesNumber) {
        this.seriesNumber = seriesNumber;
    }

    public final void setInstanceNumber(String instanceNumber) {
        this.instanceNumber = instanceNumber;
    }

    public final void setKeyObjectDescription(String keyObjectDescription) {
        this.keyObjectDescription = keyObjectDescription;
    }

    public void setRetrieveAET(String retrieveAET) {
        this.retrieveAET = retrieveAET;
    }

    public void setRetrieveURL(String retrieveURL) {
        this.retrieveURL = retrieveURL;
    }

    public void setLocationUID(String locationUID) {
        this.locationUID = locationUID;
    }
    
    public void setInstitutionName(String institutionName) {
		this.institutionName = institutionName;
	}
    
	public void setDeviceSerialNumber(String deviceSerialNumber) {
		this.deviceSerialNumber = deviceSerialNumber;
	}

	public void setTempLocation(String tempLocation) {
		this.tempLocation = tempLocation;
	}

	public void setDcmLocation(String dcmLocation) {
		this.dcmLocation = dcmLocation;
	}

	public final void setCodes(Properties codes) {
        this.codes = codes;
    }

    public final void setDocumentTitle(Attributes codeItem) {
        this.documentTitle = codeItem;
    }

    public final void setDocumentTitleModifier(Attributes codeItem) {
        this.documentTitleModifier = codeItem;
    }

    @SuppressWarnings("unchecked")
    public static void main(String[] args) throws Exception {
        try {
        	ImageIO.scanForPlugins();
        	String str = "--title DCM-113030 --institution-insname DENEME^^^SKRS12345^^^^FK12345 "
					+ "--device-sn WC001AOY "
        			+ "--location-uid 1.3.6.1.4.1.21367.2017.10.26.111 "
        			+ "--temp-tlocation C:\\STORAGE\\TEMP\\ "
        			+ "--dcm-dcmlocation C:\\Apps\\dcm4che-5.13.1\\bin "
        			+ "-o C:/Apps/btkos.dcm F:/desktop/BTHASTABIR";
        	args = str.split(" ");
            CommandLine cl = parseComandLine(args);
            final MakeKOS main = new MakeKOS();
            configure(main, cl);
            System.out.println(rb.getString("scanning"));
            DicomFiles.scan(cl.getArgList(), new DicomFiles.Callback() {
                
                @Override 
                public boolean dicomFile(File f, Attributes fmi,
                        long dsPos, Attributes ds) {
                    return main.addInstance(ds, f, tempLocation, dcmLocation);
                }
            });
            System.out.println();
            main.writeKOS();
            System.out.println(
                    MessageFormat.format(rb.getString("stored"), main.fname));
        } catch (ParseException e) {
            System.err.println("makekos: " + e.getMessage());
            System.err.println(rb.getString("try"));
            System.exit(2);
        }
    }
    
	public void dcm2Kos(String inputUrl, String outputUrl, String retrieveLocationUID, 
			String tmpRoot, String dcmRoot) throws Exception {
		String str = "--title DCM-113030 --location-uid " + retrieveLocationUID + " -o " + outputUrl + " " + inputUrl;
		CommandLine cl = parseComandLine(str.split(" "));
		final MakeKOS main = new MakeKOS();
        configure(main, cl);
        System.out.println(rb.getString("scanning"));
        DicomFiles.scan(cl.getArgList(), new DicomFiles.Callback() {
            
            @Override
            public boolean dicomFile(File f, Attributes fmi,
                    long dsPos, Attributes ds) {
                return main.addInstance(ds, f, tmpRoot, dcmRoot);
            }
        });
        System.out.println();
        main.writeKOS();
	}

    private static CommandLine parseComandLine(String[] args)
            throws ParseException{
        Options opts = new Options();
        CLIUtils.addCommonOptions(opts);
        addOptions(opts);
        CommandLine cl = CLIUtils.parseComandLine(args, opts, rb, MakeKOS.class);
        if (cl.getArgList().isEmpty())
            throw new ParseException(rb.getString("missing"));
        return cl;
    }

    @SuppressWarnings("static-access")
    public static void addOptions(Options opts) {
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("code")
                .withDescription(rb.getString("title"))
                .withLongOpt("title")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("code")
                .withDescription(rb.getString("modifier"))
                .withLongOpt("modifier")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("file|url")
                .withDescription(rb.getString("code-config"))
                .withLongOpt("code-config")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("text")
                .withDescription(rb.getString("desc"))
                .withLongOpt("desc")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("aet")
                .withDescription(rb.getString("retrieve-aet"))
                .withLongOpt("retrieve-aet")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("url")
                .withDescription(rb.getString("retrieve-url"))
                .withLongOpt("retrieve-url")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("uid")
                .withDescription(rb.getString("location-uid"))
                .withLongOpt("location-uid")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("insname")
                .withDescription(rb.getString("institution-insname"))
                .withLongOpt("institution-insname")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("sn")
                .withDescription(rb.getString("device-sn"))
                .withLongOpt("device-sn")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("tlocation")
                .withDescription(rb.getString("temp-tlocation"))
                .withLongOpt("temp-tlocation")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("dcmlocation")
                .withDescription(rb.getString("dcm-dcmlocation"))
                .withLongOpt("dcm-dcmlocation")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("no")
                .withDescription(rb.getString("series-no"))
                .withLongOpt("series-no")
                .create());
        opts.addOption(OptionBuilder
                .hasArg()
                .withArgName("no")
                .withDescription(rb.getString("inst-no"))
                .withLongOpt("inst-no")
                .create());
       opts.addOption(OptionBuilder
               .hasArg()
               .withArgName("file")
               .withDescription(rb.getString("o-file"))
               .create("o"));
       OptionGroup group = new OptionGroup();
       group.addOption(OptionBuilder
               .withLongOpt("no-fmi")
               .withDescription(rb.getString("no-fmi"))
               .create("F"));
       group.addOption(OptionBuilder
               .withLongOpt("transfer-syntax")
               .hasArg()
               .withArgName("uid")
               .withDescription(rb.getString("transfer-syntax"))
               .create("t"));
       opts.addOptionGroup(group);
       opts.addOption(OptionBuilder
               .hasArgs()
               .withArgName("[seq/]attr=value")
               .withValueSeparator('=')
               .withDescription(rb.getString("set"))
               .create("s"));
       opts.addOption(OptionBuilder
               .hasArg()
               .withArgName("suffix")
               .withDescription(rb.getString("uid-suffix"))
               .withLongOpt("uid-suffix")
               .create(null));
       CLIUtils.addEncodingOptions(opts);
   }

    private static void configure(MakeKOS main, CommandLine cl) throws Exception {
        main.setCodes(CLIUtils.loadProperties(
                cl.getOptionValue("code-config", "resource:code.properties"),
                null));
        main.setDocumentTitle(main.toCodeItem(documentTitleOf(cl)));
        if (cl.hasOption("modifier"))
            main.setDocumentTitleModifier(
                    main.toCodeItem(cl.getOptionValue("modifier")));
        main.setKeyObjectDescription(cl.getOptionValue("desc"));
        main.setRetrieveAET(cl.getOptionValue("retrieve-aet", null));
        main.setRetrieveURL(cl.getOptionValue("retrieve-url", null));
        main.setLocationUID(cl.getOptionValue("location-uid", null));
        main.setInstitutionName(cl.getOptionValue("institution-insname", null));
        main.setDeviceSerialNumber(cl.getOptionValue("device-sn", null));
        main.setTempLocation(cl.getOptionValue("temp-tlocation", "C:\\STORAGE\\TEMP\\"));
        main.setDcmLocation(cl.getOptionValue("dcm-dcmlocation", "C:\\Apps\\dcm4che-5.13.1\\bin"));
        main.setSeriesNumber(cl.getOptionValue("series-no", "999"));
        main.setInstanceNumber(cl.getOptionValue("inst-no", "1"));
        main.setOutputFile(outputFileOf(cl));
        main.setNoFileMetaInformation(cl.hasOption("F"));
        main.setTransferSyntax(cl.getOptionValue("t", UID.ExplicitVRLittleEndian));
        main.setEncodingOptions(CLIUtils.encodingOptionsOf(cl));
        CLIUtils.addAttributes(main.attrs, cl.getOptionValues("s"));
        main.setUIDSuffix(cl.getOptionValue("uid-suffix"));
    }

    public static String outputFileOf(CommandLine cl) throws MissingOptionException {
        if (!cl.hasOption("o"))
            throw new MissingOptionException(rb.getString("missing-o-file"));
        return cl.getOptionValue("o");
    }

    private static String documentTitleOf(CommandLine cl) throws MissingOptionException {
        if (!cl.hasOption("title"))
            throw new MissingOptionException(rb.getString("missing-title"));
        return cl.getOptionValue("title");
    }

    public Attributes toCodeItem(String codeValue) {
        if (codes == null)
            throw new IllegalStateException("codes not initialized");
        String codeMeaning = codes.getProperty(codeValue);
        if (codeMeaning == null)
            throw new IllegalArgumentException("undefined code value: "
                        + codeValue);
        int endDesignator = codeValue.indexOf('-');
        Attributes attrs = new Attributes(3);
        attrs.setString(Tag.CodeValue, VR.SH,
                endDesignator >= 0
                    ? codeValue.substring(endDesignator + 1)
                    : codeValue);
        attrs.setString(Tag.CodingSchemeDesignator, VR.SH,
                endDesignator >= 0
                    ? codeValue.substring(0, endDesignator)
                    : "DCM");
        attrs.setString(Tag.CodeMeaning, VR.LO, codeMeaning);
        return attrs;
    }

    public boolean addInstance(Attributes inst, File f, String tmpRoot, String dcmRoot) {
        CLIUtils.updateAttributes(inst, attrs, uidSuffix);
        String studyIUID = inst.getString(Tag.StudyInstanceUID);
        String seriesIUID = inst.getString(Tag.SeriesInstanceUID);
        String iuid = inst.getString(Tag.SOPInstanceUID);
        String cuid = inst.getString(Tag.SOPClassUID);
        String instanceNumber = inst.getString(Tag.InstanceNumber);
        int numFrame = inst.getInt(Tag.NumberOfFrames, 0);
        String modality = inst.getString(Tag.Modality);
        String seriesDescription = inst.getString(Tag.SeriesDescription);
        if (studyIUID == null || seriesIUID == null || iuid == null || cuid == null)
            return false;
        if (kos == null){
            kos = createKOS(inst, studyIUID);
        }
        if(kos != null && institutionName != null && !institutionName.equalsIgnoreCase("")){
        	kos.setString(Tag.InstitutionName, VR.LO, institutionName);
        }
        if(kos != null && deviceSerialNumber != null && !deviceSerialNumber.equalsIgnoreCase("")){
        	kos.setString(Tag.DeviceSerialNumber, VR.LO, deviceSerialNumber);
        }
        if(instanceNumber == null){
        	instanceNumber = "";
        }
        if(deviceSerialNumber == null){
        	deviceSerialNumber = "";
        }
        refSOPSeq(refSeriesSeq(studyIUID), seriesIUID, 
        		modality, seriesDescription, f, tmpRoot, dcmRoot).add(refSOP(cuid, iuid, instanceNumber, numFrame));
        contentSeq.add(contentItem(valueTypeOf(inst), refSOP(cuid, iuid, instanceNumber, numFrame)));
        return true;
    }

    public void writeKOS() throws IOException {
        DicomOutputStream dos = new DicomOutputStream(
                new BufferedOutputStream(fname != null 
                        ? new FileOutputStream(fname)
                        : new FileOutputStream(FileDescriptor.out)),
                nofmi ? UID.ImplicitVRLittleEndian
                      : UID.ExplicitVRLittleEndian);
        dos.setEncodingOptions(encOpts);
        try {
            dos.writeDataset(
                    nofmi ? null : kos.createFileMetaInformation(tsuid),
                    kos);
        } finally {
            dos.close();
        }
    }

    private Sequence refSeriesSeq(String studyIUID) {
        for (Attributes refStudy : evidenceSeq)
            if (studyIUID.equals(refStudy.getString(Tag.StudyInstanceUID)))
                return refStudy.getSequence(Tag.ReferencedSeriesSequence);

        Attributes refStudy = new Attributes(2);
        Sequence refSeriesSeq = refStudy.newSequence(Tag.ReferencedSeriesSequence, 10);
        refStudy.setString(Tag.StudyInstanceUID, VR.UI, studyIUID);
        evidenceSeq.add(refStudy);
        return refSeriesSeq;
    }

    private Sequence refSOPSeq(Sequence refSeriesSeq , String seriesIUID, 
    		String modality, String seriesDescription, File f, String tmpRoot, String dcmRoot) {
        for (Attributes refSeries : refSeriesSeq)
            if (seriesIUID.equals(refSeries.getString(Tag.SeriesInstanceUID)))
                return refSeries.getSequence(Tag.ReferencedSOPSequence);
 
        Attributes refSeries = new Attributes(6);
        if (retrieveAET != null)
            refSeries.setString(Tag.RetrieveAETitle, VR.AE, retrieveAET);
        if (retrieveURL != null)
            refSeries.setString(Tag.RetrieveURL, VR.UR, retrieveURL);
        Sequence refSOPSeq = refSeries.newSequence(Tag.ReferencedSOPSequence, 100);
        
        Attributes iconImg = new Attributes(7);
        iconImg.setInt(Tag.Rows, VR.US, 64);
        iconImg.setInt(Tag.Columns, VR.US, 64);
        iconImg.setInt(Tag.BitsAllocated, VR.US, 8);
        iconImg.setInt(Tag.BitsStored, VR.US, 8);
        iconImg.setInt(Tag.HighBit, VR.US, 7);
        iconImg.setInt(Tag.SamplesPerPixel, VR.US, 3);
        iconImg.setString(Tag.PhotometricInterpretation, VR.CS, "YBR_FULL_422");
        iconImg.setString(Tag.PixelAspectRatio, VR.IS, "1");
        try {
        	Attributes tmpAttr = parse(f);
        	if(tmpAttr.getBytes(Tag.PixelData) != null){
        		Sequence iconImgSeq = refSeries.newSequence(Tag.IconImageSequence, 1);
        		iconImg.setBytes(Tag.PixelData, VR.OB, addThumbnailBytes(tmpRoot, dcmRoot, f));
        		iconImgSeq.add(iconImg);
        	}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        refSeries.setString(Tag.SeriesInstanceUID, VR.UI, seriesIUID);
        refSeries.setString(Tag.SeriesDescription, VR.UI, seriesDescription);
        refSeries.setString(Tag.Modality, VR.CS, modality);
        if (locationUID != null)
            refSeries.setString(Tag.RetrieveLocationUID, VR.UI, locationUID);
        refSeriesSeq.add(refSeries);
        return refSOPSeq;
    }
    
    private static Attributes parse(File file) throws IOException {
        DicomInputStream in = new DicomInputStream(file);
        try {
            in.setIncludeBulkData(IncludeBulkData.NO);
            return in.readDataset(-1, -1);
        } finally {
            SafeClose.close(in);
        }
    }
    
    public String addThumbnail(String tmpRoot, String dcmRoot, File f) throws Exception{
    	String tempThumbnailUrl = dcm2jpg(tmpRoot, f.getAbsolutePath(), dcmRoot);
		BufferedImage img = ImageIO.read(new File(tempThumbnailUrl));
		BufferedImage bufferedImage = Scalr.resize(img, 64);
		String thumbnailImage = encodeBufferedImageToBase64Binary(bufferedImage); 
		return thumbnailImage;
    }
    
    private static String encodeBufferedImageToBase64Binary(BufferedImage bufferedImage)
			throws IOException { 
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ImageIO.write( bufferedImage, "jpg", baos );
		baos.flush();
		byte[] bytes = baos.toByteArray();
		baos.close();
		byte[] encoded = java.util.Base64.getEncoder().encode(bytes);
		String encodedString = new String(encoded);
		return encodedString;
	}
    
    public byte[] addThumbnailBytes(String tmpRoot, String dcmRoot, File f) throws Exception{
    	String tempThumbnailUrl = dcm2jpg(tmpRoot, f.getAbsolutePath(), dcmRoot);
		BufferedImage img = ImageIO.read(new File(tempThumbnailUrl));
		BufferedImage bufferedImage = Scalr.resize(img, 64);
		byte[] thumbnailImageBytes = bufferedImageToBytes(bufferedImage); 
		return thumbnailImageBytes;
    }
    
    public static byte[] bufferedImageToBytes(BufferedImage bufferedImage) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ImageIO.write(bufferedImage, "jpg", baos );
		baos.flush();
		byte[] imageInByte = baos.toByteArray();
		baos.close();
		return imageInByte;
	}
    
    public static String dcm2jpg(String tmpRoot, String fileName, String dcmRoot) throws IOException, InterruptedException {
    	String tempUrl = tmpRoot + RandomGUID.getRandomGUID() + ".jpg";
    	try
    	{ 
    		Runtime rt = Runtime.getRuntime();
    		Process p = rt.exec(new String[]{dcmRoot + "/dcm2jpg.bat", fileName, tempUrl});
    		p.waitFor();
    		/*
    		InputStream stderr = p.getErrorStream();
    		InputStreamReader isr = new InputStreamReader(stderr);
    		BufferedReader br = new BufferedReader(isr);
    		String line = null;
    		System.out.println("<ERROR>");
    		while ( (line = br.readLine()) != null)
    			System.out.println(line);
    		System.out.println("</ERROR>");
    		int exitVal = p.waitFor();
    		System.out.println("Process exitValue: " + exitVal);
    		*/
    	} catch (Throwable t){
    		t.printStackTrace();
    	}
    	return tempUrl;
    }
    
    private String valueTypeOf(Attributes inst) {
        return inst.contains(Tag.PhotometricInterpretation) ? "IMAGE"
                      : inst.contains(Tag.WaveformSequence) ? "WAVEFORM"
                                                            : "COMPOSITE";
    }

    private Attributes refSOP(String cuid, String iuid, String instanceNumber, int numFrame) {
        Attributes item = new Attributes(4);
        item.setString(Tag.ReferencedSOPClassUID, VR.UI, cuid);
        item.setString(Tag.ReferencedSOPInstanceUID, VR.UI, iuid);
        item.setString(Tag.InstanceNumber, VR.IS, instanceNumber);
        item.setInt(Tag.NumberOfFrames, VR.IS, numFrame);
        return item;
    }

    private Attributes createKOS(Attributes inst, String studyUid) {
        Attributes attrs = new Attributes(inst, PATIENT_AND_STUDY_ATTRS);
        attrs.setString(Tag.InstitutionName, VR.LO, inst.getString(Tag.InstitutionName));
        attrs.setString(Tag.DeviceSerialNumber, VR.LO, inst.getString(Tag.DeviceSerialNumber));
        attrs.setString(Tag.PatientBirthDate, VR.DA, inst.getString(Tag.PatientBirthDate));
        attrs.setString(Tag.PatientSex, VR.CS, inst.getString(Tag.PatientSex));
        attrs.setString(Tag.SOPClassUID, VR.UI, UID.KeyObjectSelectionDocumentStorage);
        attrs.setString(Tag.SOPInstanceUID, VR.UI, UIDUtils.createUID());
        attrs.setDate(Tag.ContentDateAndTime, new Date());
        attrs.setString(Tag.Modality, VR.CS, "KO");
        attrs.setNull(Tag.ReferencedPerformedProcedureStepSequence, VR.SQ);
        attrs.setString(Tag.StudyInstanceUID, VR.UI, studyUid);
        attrs.setString(Tag.SeriesInstanceUID, VR.UI, UIDUtils.createUID());
        attrs.setString(Tag.SeriesNumber, VR.IS, seriesNumber);
        attrs.setString(Tag.InstanceNumber, VR.IS, instanceNumber);
        attrs.setString(Tag.ValueType, VR.CS, "CONTAINER");
        attrs.setString(Tag.ContinuityOfContent, VR.CS, "SEPARATE");
        attrs.newSequence(Tag.ConceptNameCodeSequence, 1).add(documentTitle);
        evidenceSeq = attrs.newSequence(Tag.CurrentRequestedProcedureEvidenceSequence, 1);
        attrs.newSequence(Tag.ContentTemplateSequence, 1).add(templateIdentifier());
        contentSeq = attrs.newSequence(Tag.ContentSequence, 1);
        if (documentTitleModifier != null)
            contentSeq.add(documentTitleModifier());
        if (keyObjectDescription != null)
            contentSeq.add(keyObjectDescription());
        return attrs;
    }

    private Attributes templateIdentifier() {
        Attributes attrs = new Attributes(2);
        attrs.setString(Tag.MappingResource, VR.CS, "DCMR");
        attrs.setString(Tag.TemplateIdentifier, VR.CS, "2010");
        return attrs ;
    }

    private Attributes documentTitleModifier() {
        Attributes item = new Attributes(4);
        item.setString(Tag.RelationshipType, VR.CS, "HAS CONCEPT MOD");
        item.setString(Tag.ValueType, VR.CS, "CODE");
        item.newSequence(Tag.ConceptNameCodeSequence, 1).add(toCodeItem("DCM-113011"));
        item.newSequence(Tag.ConceptCodeSequence, 1).add(documentTitleModifier);
        return item;
    }

    private Attributes keyObjectDescription() {
        Attributes item = new Attributes(4);
        item.setString(Tag.RelationshipType, VR.CS, "CONTAINS");
        item.setString(Tag.ValueType, VR.CS, "TEXT");
        item.newSequence(Tag.ConceptNameCodeSequence, 1).add(toCodeItem("DCM-113012"));
        item.setString(Tag.TextValue, VR.UT, keyObjectDescription);
        return item;
    }

    private Attributes contentItem(String valueType, Attributes refSOP) {
        Attributes item = new Attributes(3);
        item.setString(Tag.RelationshipType, VR.CS, "CONTAINS");
        item.setString(Tag.ValueType, VR.CS, valueType);
        item.newSequence(Tag.ReferencedSOPSequence, 1).add(refSOP);
        return item;
    }

}